/*
 * vRouter: DataLink
 *
 * Purpose: Implements DataLinks, which behave roughly like layer-3
 *          switches.
 *
 * Changes:
 *  0.02 - Made routers_ and ipaddresses_ vectors to allow unlimited
 *         routers on a network.
 *
 */

#include <string>
#include <iostream>
#include "ipaddress.h"
#include "router.h"
#include "packet.h"
 
#include <string.h>


/*
 * Construct a new DataLink 
 */

DataLink::DataLink(IPAddress network, int bandwidth=1000, int delay=100)
{
  
  /* routers_ and ipaddresses_ are now vectors.  No need 
     to initialize them here anymore */
     
  net_ = network; // network address
  bandwidth_ = bandwidth; // bandwidth in bps
  delay_ = delay; // delay in ms

}

/*
 * Destructor 
 */
DataLink::~DataLink()
{
  // not much to do
}

/*
 * Add a router to the datalink
 */
void DataLink::addRouter(Router *rtr, IPAddress a)
{
  
  routers_.push_back(rtr);
  ipaddresses_.push_back(new IPAddress(a));


}

/*
 * Return the network address of this datalink
 */
IPAddress DataLink::net()
{
  return net_;
}

/*
 * Print a list of routers on this datalink to standard out
 */
void DataLink::listHosts()
{
  cout << " ---------  Routers on " << net_.toStringPrefix() << " ----------" << endl;
  for(unsigned int i = 0; i < routers_.size(); i++)
    cout << routers_[i]->name() << " : " << ipaddresses_[i]->toString() << endl;

}

/*
 * Enqueue the given packet to nextHop.  The packet will
 * actually be delivered when the datalink is given time via
 * a call to runTimeSlice()
 */
void DataLink::deliverPacket(Packet *p, IPAddress nextHop)
{
  // add packet to the delivery Queue "Q"

  Q.push_back(Qentry(p, nextHop));

}
  
/*
 * Deliver all the packets in the Q
 */
void DataLink::runTimeSlice(int timeWidth)
{

  /* randomize delivery order */

  for(unsigned int i =0; i < Q.size(); i ++)
    {
      int x = (int) (Q.size()*rand()/(RAND_MAX+1.0));
      int y = (int) (Q.size()*rand()/(RAND_MAX+1.0));
     
      /* Swap the packets in slots x and y */
      Qentry t;
      t = Q[x];
      Q[x] = Q[y];
      Q[y] = t;
      
    }
  
  /* Deliver all the packets in the Q */

  /* FIXME: We don't enforce bandwidth restrictions.  This is left
     up to routers, which don't send more than can be delivered.
     This makes more than 2 routers on a datalink problematic.  
  */

  /* Compute the number of bits in the transmission queue */
  int bitsinQ =0;
  for(unsigned int i =0; i < Q.size(); i++)
    {
      bitsinQ += Q[i].p->size();
    }
  bitsinQ *= 8;

  /* Compute the number of bits that can be sent in the timeslice */
  int timeslicebits  = (int) (timeWidth * bandwidth_ * .001); // timewidth in ms

  /* compute the percent load */
  float load = (float)bitsinQ / (float)timeslicebits;


  /* Must... avoid... division... by zero! */
  if(load ==1)
    load = .999;

  float udelay ;

  
  for(unsigned int i = 0; i < Q.size(); i ++)
    {
      Packet *p = Q[i].p;
      IPAddress nxHop = Q[i].nxHop;

      p->setDelay(p->delay() + delay_);
      
      /* Use equations from queuing theory to calculate a delay factor
       * based on link load level. 
       * delay = servicetime * (2 - load) / 2*(1-load)
       */
      
      float servicetime = (float)p->size() / (float)bandwidth_;
      udelay = (servicetime* (2-load)) / (2*(1-load));

      /* usually, these delays are small, so we keep track of them on a 
       * usec level
       */
      udelay *=1000000;

      p->setUDelay(p->udelay() + udelay);

      for(unsigned int i =0; i < ipaddresses_.size(); i++)
	if(*ipaddresses_[i] == nxHop || p->dst().isBroadcast())
	  {
	    routers_[i]->deliverPacket(p);
	    break;
	  }
    }

  Q.clear();
}





